@@ -2,9 +2,9 @@
 
 # egd.pl: The Entropy Gathering Daemon
 #  by Brian Warner <warner@lothar.com>
-#  Released Version 0.6
+#  Released Version 0.8
 #
-#  $Id: egd.pl,v 1.44 2000/04/03 06:49:39 warner Exp $
+#  $Id: egd.pl,v 1.48 2000/05/14 01:50:17 warner Exp $
 
 # This is a perl program to collect entropy from a running Unix-like system,
 # mostly through various stats calls (vmstat and friends). The output is
@@ -22,7 +22,7 @@
 # Cryptlib, by Peter Gutmann et al. The basic list of entropy sources used
 # here was copied from that code. Everything else is original.
 
-# Copyright (C) 1999, Brian Warner <warner@lothar.com>
+# Copyright (C) 1999,2000 Brian Warner <warner@lothar.com>
 # Released under the GNU General Public License
 # For details, see the file COPYING provided in the source archive
 
@@ -406,7 +406,6 @@ sub add_entropy {
     my($data, $entropy) = @_;
     $pool_entropy += $entropy;
     # stir in data
-    my $c = new SHA;
     while(length($data)) {
 	my $chunk = substr($data, 0, $DIGEST_SIZE);
 	$pool ^= $chunk;
@@ -449,6 +448,8 @@ sub stir {
 
 sub shutdown_thing {
     my($c) = @_;
+    return unless $c->{'active'}; # skip duplicate shutdowns
+    $c->{'active'} = 0;
     print STDERR "shutdown_thing($c)\n" if $debug_select;
     if ($c->{'type'} eq 'gatherer') {
 	print STDERR "reaping pid ",$c->{'pid'}," source ",$c->{'source'},"\n"
@@ -641,6 +642,12 @@ while (1) {
 		    }
 		    $next_source = ($next_source +1) % @sources;
 		}
+		# If you are running EGD without any gatherers (i.e. just as
+		# an entropy repository for data written in by clients
+		# with the 0x03 write command), then remove this die(),
+		# because you're ok with the lack of sources
+		die "all sources are unusable. aborting.."
+		  unless defined($next_available);
 		print STDERR " next available is in ",
 		  ($next_available - time()),"\n" if $debug_gather;
 		goto done_spawn;
@@ -652,6 +659,7 @@ while (1) {
 		  # 'pid' filled in by fork
 		  'expire' => time() + $sources[$s]->{'timeout'},
 		  'type' => 'gatherer',
+		  'active' => 1,
 		};
 	$sources[$s]->{'inuse'} = 1;
 	$sources[$s]->{'last_started'} = time;
@@ -744,17 +752,6 @@ while (1) {
 	$refresh_time = time() + $REFRESH_TIME
     }
 
-    # first shut down the sockets for folks that have hung up
-    foreach my $s (@$errorful) {
-	if (grep {$s->[1] == $_} @lsockets) {	# is == the right test?
-	    die "problem: lsocket had error\n";
-	}
-	# shut it down, kill it off, reap it
-	print STDERR "shutting down errorful socket $s\n" if $debug_select;
-	my $c = $s->[1];
-	shutdown_thing($c);
-    }
-
     # then push out all the text we can
     foreach my $s (@$writable) {
 	my $c = $s->[1];
@@ -788,7 +785,7 @@ while (1) {
 		print STDERR "new client from $peer on lsocket $l->[0]\n";
 	    }
 	    my $new_client = { 'fd' => $new, 'type' => 'client',
-			       'canwrite' => $l->[1]};
+			       'canwrite' => $l->[1], 'active' => 1 };
 	    push(@clients, $new_client);
 	    $read_select->add([$new, $new_client]);
 	    $error_select->add([$new, $new_client]);
@@ -812,6 +809,18 @@ while (1) {
 	}
     }
 
+    # now shut down the sockets for folks that have hung up, or gatherers
+    # that have died
+    foreach my $s (@$errorful) {
+	if (grep {$s->[1] == $_} @lsockets) {	# is == the right test?
+	    die "problem: lsocket had error\n";
+	}
+	# shut it down, kill it off, reap it
+	print STDERR "shutting down errorful socket $s\n" if $debug_select;
+	my $c = $s->[1];
+	shutdown_thing($c);
+    }
+
     # check to see if any gatherer ought to be expired
     my $now = time();
     my(@old) = grep { $_->{'expire'} < $now } @gatherers;
@@ -923,6 +932,11 @@ advice from pgutmann library:
  #add a 'get pid' command
  #add a --kill arg that gets a pid and kills that process
 
- then for selftests, can run with --bottomless --quitafter 30 at start,
-  with --kill at end of tests. If desired, gpg can run --quitafter 30
-  automatically when it starts,
+ #then for selftests, can run with --bottomless --quitafter 30 at start,
+ # with --kill at end of tests. If desired, gpg can run --quitafter 30
+ # automatically when it starts,
+
+ #solaris-2.8, Irix 6.5.6: all gatherers fail?
+ # select() marks pipes at EOF both readable and errorful, and we dropped the
+ # errorful ones first. Switch to reap EOF before checking errorful ones,
+ # don't mark gatherer unusable in that case.
